""" Basic unit tests for axographio

This module contains the tests (and doctest wrappers) for the axographio
module.
"""

import doctest
import inspect
import unittest
import numpy as np
import os.path
import pkg_resources
import os
import tempfile
import copy

import axographio

example_files = {
    'old_digitized_format': pkg_resources.resource_filename(__name__,
        '../include/axograph_readwrite/AxoGraph Digitized File'),
    'old_graph_format':     pkg_resources.resource_filename(__name__,
        '../include/axograph_readwrite/AxoGraph Graph File'),
    'axograph_x_format':    pkg_resources.resource_filename(__name__,
        '../include/axograph_readwrite/AxoGraph X File.axgx'),
    }

class TestSampleFiles(unittest.TestCase):
    """Test the ability to read some sample data files.

    These files were provided by Axograph with their sample code in
    Axograph X, and generously released into the public domain.
    """


    def test_digitizedfile(self):
        file = axographio.read(example_files['old_digitized_format'])
        # do some sanity checks to make sure the file loaded as expected
        self.assertEqual(file.fileformat, axographio.old_digitized_format)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(len(file.names), 29)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(file.names[0], 'Time (s)')
        self.assertEqual(file.names[1], 'Current (A)')
        self.assertEqual(file.names[28], 'Column29')
        for d in file.data:
            self.assertEqual(len(d), 200)
        self.assertEqual(round(sum(file.data[1]),13), -1.973206e-07)
        self.assertEqual(round(sum(file.data[28]),7), -0.0101875)
        self.assertEqual(round(file.data[0][199],6), 0.02)


    def test_graphfile(self):
        file = axographio.read(example_files['old_graph_format'])
        # do some sanity checks to make sure the file loaded as expected
        self.assertEqual(file.fileformat, axographio.old_graph_format)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(len(file.names), 3)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(file.names[0], 'Time (sec)')
        self.assertEqual(file.names[1], 'Current (A)')
        self.assertEqual(file.names[2], 'Current (A)')
        for d in file.data:
            self.assertEqual(len(d), 2048)
        self.assertEqual(round(sum(file.data[1]),13), -2.1483816e-6)
        self.assertEqual(round(sum(file.data[2]),13), -1.4318602e-6)
        self.assertEqual(round(float(file.data[0][2047]),6), 1.024)


    def test_axograph_x_file(self):
        file = axographio.read(example_files['axograph_x_format'])
        # do some sanity checks to make sure the file loaded as expected
        self.assertEqual(file.fileformat, axographio.axograph_x_format)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(len(file.names), 7)
        self.assertEqual(len(file.names), len(file.data))
        self.assertEqual(file.names[0], 'Time (s)')
        self.assertEqual(file.names[1], 'Current (A)')
        self.assertEqual(file.names[2], '')
        self.assertEqual(file.names[6], '')
        for d in file.data:
            self.assertEqual(len(d), 1000)
        self.assertEqual(round(sum(file.data[1]),15), -2.3021573e-8)
        self.assertEqual(round(sum(file.data[6]),15), -2.5514375e-8)
        self.assertEqual(file.data[0][999], 0.05)



class TestReadWrite(unittest.TestCase):
    """Test reading and writing of different data types"""

    def roughly(self, x, accuracy):
        # round x to about 'accuracy' digits of precision
        array = np.asarray(x, dtype=np.float64)

        scale = np.exp(round(np.log(array.max())))
        return ((array/scale).round(accuracy))*scale

    def testreadwrite(self):
        # create a file with all of the supported data types
        timecol = [2.01, 2.02, 2.03, 2.04]
        listcol = [1.0, 2.1, 3.4, 5.3]
        seqcol = axographio.linearsequence(4, 0.12345, 0.54321)
        scaledcol = axographio.scaledarray([12, -18, 16, 42], 0.001, 0.1)
        floatcol = np.array([1.8218, 4.3672, 5.1484, 7.3235], dtype=np.float32)
        doublecol = np.array(
                [1.82187845, 4.367244252, 5.14842452445, 7.322452435],
                dtype=np.float64)
        shortcol = np.array([11043, -21274, -1834, 32341], dtype=np.int16)
        intcol = np.array([1213104371, -2027483727, -18, 3234], dtype=np.int32)
        originalfile = axographio.file_contents(
                ['time', '', 'seq', 'float', 'double', 'short', 'int'],
                [timecol, listcol, seqcol, floatcol, doublecol, shortcol,
                    intcol]
                )

        # try rewriting and rereading it multiple times
        for fileformat in axographio.supported_formats:
            currentfile = copy.deepcopy(originalfile) # to protect the original
            currentfile.fileformat = fileformat

            # the older formats may lose precision; that's OK for this
            # test.
            if fileformat == axographio.old_graph_format:
                accuracy = 5 # digits
            elif fileformat == axographio.old_digitized_format:
                accuracy = 3 # digits
            else:
                accuracy = 16 # digits

            # Do more than one pass of reading and writing to make sure the
            # data remains consistent.
            for pass_number in range(2):
                handle, tempfilename = tempfile.mkstemp()
                try:
                    currentfile.write(tempfilename)
                    currentfile = axographio.read(tempfilename)
                finally:
                    os.close(handle)
                    os.remove(tempfilename)

                # make sure the newly read data matches the original version
                self.assertEqual(currentfile.fileformat, fileformat)
                self.assertEqual(len(originalfile.names),
                        len(currentfile.names))
                self.assertEqual(len(originalfile.data), len(currentfile.data))
                self.assertTrue(np.all([a == b
                    for a,b in zip(originalfile.names, currentfile.names)]))

                for a,b in zip(originalfile.data, currentfile.data):
                    self.assertTrue(np.all(
                        self.roughly(a, accuracy) == self.roughly(b, accuracy)))



class TestRegressions(unittest.TestCase):
    """Tests for bugs that were found in previous releases"""

    # bugs fixed in 0.1.1
    def test_linearsequencelength(self):
        """ Off-by-one error in len(np.asarray(a_linearsequence))
        Found by: Dr. Hillel Chiel (hjc@case.edu)
        """
        seq = axographio.linearsequence(1000, 0., 0.036)
        seqAsArray = np.asarray(seq)
        self.assertEqual(len(seq), 1000)
        self.assertEqual(len(seqAsArray), 1000)



def test_suite():
    """Returns a test suite with all of the tests for axographio"""
    suite = unittest.TestSuite()
    fix_module_doctest(axographio.extension)
    suite.addTest(doctest.DocTestSuite(axographio.extension))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestSampleFiles))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestReadWrite))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestRegressions))
    return suite

def run():
    """Run the test suite containing all of the tests for axographio"""

    # doctests need the array formatting used in numpy < 1.14
    try:
        np.set_printoptions(legacy='1.13')
    except TypeError:
        pass

    unittest.TextTestRunner(verbosity=2).run(test_suite())

#
# The following two functions work around an incompatability with doctest and
# Cython.  These two functions were cut and pasted from the Cython FAQ at
# https://github.com/cython/cython/wiki/FAQ#how-can-i-run-doctests-in-cython-code-pyx-files
# - see this FAQ for more details.
#


def _from_module(module, object):
    """ Return true if the given object is defined in the given module."""
    if module is None:
        return True
    elif inspect.getmodule(object) is not None:
        return module is inspect.getmodule(object)
    elif inspect.isfunction(object):
        return module.__dict__ is object.func_globals
    elif inspect.isclass(object):
        return module.__name__ == object.__module__
    elif hasattr(object, '__module__'):
        return module.__name__ == object.__module__
    elif isinstance(object, property):
        return True # [XX] no way not be sure.
    else:
        raise ValueError("object must be a class or function")

def fix_module_doctest(module):
    """
    Extract docstrings from cython functions, that would be skipped by doctest
    otherwise.
    """
    module.__test__ = {}
    for name in dir(module):
       value = getattr(module, name)
       if (inspect.isbuiltin(value) and isinstance(value.__doc__, str)
               and _from_module(module, value)):
           module.__test__[name] = value.__doc__


if __name__ == '__main__':
    run()



"""Test suite for axographio

Run all tests using axographio.tests.run()
"""

import unittest
from . import test_axographio
from .test_axographio import run

# explicit listing needed to ensure pydoc/help() finds everything
__all__ = [
    'run',
    ]



# cython: c_string_type=str, c_string_encoding=ascii

import numpy as np
cimport numpy as np

cdef extern from "stdlib.h":
    ctypedef int size_t
    void* malloc(size_t num)
    void free(void*)
    void* memcpy(void* destination, void* source, size_t num)
    void* memset(void* destination, int source, size_t num)

cdef extern from "include/axograph_readwrite/fileUtils.h":
    ctypedef void* AGDataRef
    ctypedef extern char* const_char_ptr "const char*"
    AGDataRef NewFile( const_char_ptr fileName )
    AGDataRef OpenFile( const_char_ptr fileName )
    void CloseFile( AGDataRef dataRefNum )


cdef extern from "include/axograph_readwrite/AxoGraph_ReadWrite.h":
    ctypedef int int32_t
    enum ag_errors:
        kAG_MemoryErr, kAG_FormatErr, kAG_VersionErr

    enum ColumnType:
        IntType,
        DoubleType,
        BoolType,
        StringType,
        ShortArrayType,
        IntArrayType,
        FloatArrayType,
        DoubleArrayType,
        BoolArrayType,
        SeriesArrayType,
        ScaledShortArrayType,
        StringArrayType,
        ReferenceType

    struct SeriesArray:
        double firstValue
        double increment


    struct ScaledShortArray:
        double scale
        double offset
        short *shortArray

    struct ColumnData:
        ColumnType type
        int32_t points
        int32_t titleLength
        unsigned char *title
        # the rest of this structure is a union:
        short *shortArray
        int32_t *intArray
        float *floatArray
        double *doubleArray
        SeriesArray seriesArray
        ScaledShortArray scaledShortArray

    int kAxoGraph_Graph_Format
    int kAxoGraph_Digitized_Format
    int kAxoGraph_X_Format

    int AG_GetFileFormat( AGDataRef refNum, int *fileFormat )

    int AG_GetNumberOfColumns( AGDataRef refNum, int fileFormat,
            int32_t *numberOfColumns )

    int AG_ReadColumn( AGDataRef refNum, int fileFormat,
            int columnNumber, ColumnData *columnData )

    int AG_ReadFloatColumn( AGDataRef refNum, int fileFormat,
            int columnNumber, ColumnData *columnData )

    int AG_WriteHeader( AGDataRef refNum, int fileFormat, int numColumns )

    int AG_WriteColumn( AGDataRef refNum, int fileFormat,
            int columnNumber, ColumnData *columnData )



# supported file formats
old_graph_format = kAxoGraph_Graph_Format #: pre-Axograph X graph format
old_digitized_format = kAxoGraph_Digitized_Format #: pre-Axograph X format
axograph_x_format = kAxoGraph_X_Format #: Axograph X format

newest_format = axograph_x_format #: the most current format supported
supported_formats = [old_graph_format, old_digitized_format, axograph_x_format]

class file_contents:
    """The contents of an axograph data file

    names is a list of column names.

    data contains the contents of each column.

    fileformat is the file format to use; currently old_graph_format,
        old_digitized_format, and axograph_x_format are supported,
        with newest_format being a synonym for the most recent format
        supported (currently axograph_x_format).

    """
    def __init__(self, names, data, fileformat = newest_format):
        self.names = names
        self.data = data
        self.fileformat = fileformat

    def write(self, filename):
        """Write this file to the given filename"""
        cdef int result
        cdef ColumnData columndata

        # open the file
        cdef AGDataRef file = NewFile(filename)
        if file == NULL:
            raise IOError('file not found')

        try:

            # write file the header
            result = AG_WriteHeader( file, self.fileformat, len(self.data) )
            if result != 0:
                raise IOError((result,
                    'AG_WriteHeader returned error %d' % result))

            # write each column
            for i in range(len(self.data)):
                prepare_columndata(&columndata, i, self.fileformat,
                        self.names[i], self.data[i])

                result = AG_WriteColumn(file, self.fileformat, i, &columndata)
                if result != 0:
                    raise IOError((result,
                        'AG_WriteColumn returned error %d' % result))
                free_columndata(&columndata)

        finally:
            CloseFile(file)



class _getitem_iterator:
    """A simple iterator for objects that support __getitem__ and __len__

    This class provides a simple (but slow) iterator that can be added to
    classes that already support __getitem__ and __len__.

    This can be created for an object

    >>> l = [3,1,4]
    >>> i = _getitem_iterator(l)

    __iter__ will return itself, as required in the language spec

    >>> i is iter(i)
    True

    next can be used to walk through the items in the list, eventually
    raising a StopIteration exception.

    >>> i.next()
    3
    >>> i.next()
    1
    >>> i.next()
    4
    >>> i.next()
    Traceback (most recent call last):
        ...
    StopIteration

    """

    def __init__(self, array):
        self.array = array
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.array):
            raise StopIteration()
        self.index += 1
        return self.array[self.index - 1]

    next = __next__ # for Python 2



class linearsequence:
    """A sequence of terms of the form  a_k = m * k + b,   0 <= k < n

    Since samples are typically taken at a fixed frequency, Axograph
    stores the time column as just a starting time (b), number of points (n),
    and interval between points (m).  This class exposes this data as a
    sequence which can easily be converted to an array using numpy.asarray.

    >>> seq = linearsequence(5, 1.125, 0.25)
    >>> len(seq)
    5

    This supports integer indices:

    >>> seq[0]
    1.125
    >>> seq[1]
    1.375
    >>> seq[4]
    2.125
    >>> seq[-1]
    2.125
    >>> seq[-5]
    1.125

    It also works with slices:

    >>> seq[1:11:2]
    array([ 1.375,  1.875])


    Bad indices will result in an exceptions

    >>> seq[5]
    Traceback (most recent call last):
        ...
    IndexError: index out of bounds
    >>> seq[-6]
    Traceback (most recent call last):
        ...
    IndexError: index out of bounds
    >>> seq['foo']
    Traceback (most recent call last):
        ...
    TypeError: not a valid index type


    This can be converted to an array using numpy.asarray:

    >>> import numpy
    >>> numpy.asarray(seq)
    array([ 1.125,  1.375,  1.625,  1.875,  2.125])

    or iterated over:

    >>> sum(iter(seq))
    8.125

    """

    def __init__(self, numpoints, start, step):
        """ Creates a new sequence

        Creates a new linear sequence whose first value is start, and where
        each subsequent value is step larger than the value before it.
        """
        self.numpoints = numpoints
        self.start = start
        self.step = step

    def __getitem__(self, index):
        """Implements the s[n] operator"""
        if isinstance(index, int):
            if index < -self.numpoints or index >= self.numpoints:
                raise IndexError('index out of bounds')
            elif index < 0:
                return (self.numpoints + index) * self.step + self.start
            else:
                return index * self.step + self.start
        elif isinstance(index, slice):
            return (np.arange(*index.indices(self.numpoints)) * self.step
                    + self.start)
        else:
            raise TypeError('not a valid index type')

    def __len__(self):
        """Implements len(s)"""
        return self.numpoints

    def __array__(self, dtype=np.float64):
        """Implements numpy.asarray(s)"""
        return np.linspace(self.start, self.start + self.step * self.numpoints,
                self.numpoints, endpoint=False)

    def __iter__(self):
        """Implements iter(s)"""
        return _getitem_iterator(self)



def aslinearsequence(x):
    """ Convert an object to a linear sequence

    If x is a linear sequence, this function will just return x.
    Otherwise, it will try converting x to an array and checking if the
    points form a linear sequence; if so it will return the sequence,
    otherwise it will raise a TypeError.

    >>> x = [1.2, 2.3, 3.4, 4.5]
    >>> y = aslinearsequence(x)
    >>> x is y
    False
    >>> np.asarray(y)
    array([ 1.2,  2.3,  3.4,  4.5])
    >>> z = aslinearsequence(y)
    >>> y is z
    True
    >>> aslinearsequence([1.2, 2.3, 4.4, 4.5])
    Traceback (most recent call last):
        ...
    TypeError: Data could not be converted to a linear sequence

    """
    if isinstance(x, linearsequence): # nothing to do
        return x
    else:
        x = np.asarray(x)

        if len(x) == 0:
            return linearsequence(0, 0., 0.)
        elif len(x) == 1:
            return linearsequence(1, x[0], 0)
        else:
            step = (x[-1] - x[0]) / (len(x) - 1)

            diffs = (x[1:] - x[:-1]) / step - 1
            if (np.max(np.abs(diffs)) > 1e-6):
                raise TypeError(
                        'Data could not be converted to a linear sequence')

            return linearsequence(len(x), x[0], step)



class scaledarray:
    """A sequence of samples as 16 bit values with scaling and offsets

    This class represents a data format used by axograph, where each value
    is a single 16 bit int.  To convert these values to the actual
    measurements, one multiplies by an scale factor and adds an offset.

    This supports the standard sequence operations, and can be converted to
    an array using numpy.asarray.

    >>> import numpy
    >>> sa = scaledarray(numpy.array([2,7,1,8]), 0.5, 0.25)
    >>> numpy.asarray(sa)
    array([ 1.25,  3.75,  0.75,  4.25])
    >>> len(sa)
    4

    >>> sa[1]
    3.75
    >>> sum(iter(sa))
    10.0

    """
    def __init__(self, data, scale, offset):
        """Create a scaledarray

        Creates a new scaledarray where each element is equal to its
        corresponding element in data times a scaling factor plus an offset.

        """
        self.data = np.asarray(data, dtype=np.int16)
        self.scale = scale
        self.offset = offset

    def __getitem__(self, index):
        """Implements the s[n] operator"""
        return self.data[index] * self.scale + self.offset

    def __len__(self):
        """Implements len(s)"""
        return len(self.data)

    def __array__(self, dtype=np.float64):
        """Implements numpy.asarray(s)"""
        return np.asarray(self.data * self.scale + self.offset, dtype=dtype)

    def __iter__(self):
        """Implements iter(s)"""
        return _getitem_iterator(self)



def asscaledarray(x, forceoffset = None):
    """ Convert an object to a scaled array

    If x is a scaled array, this function will just return x.
    Otherwise, it will try converting x to an array of doubles and
    distributing the values over the given range.  If forceoffset is
    given it will be used as the offset; otherwise the mean will be used.

    >>> x = [1.25, 8.125, 6.75, 4.625, 9.25]
    >>> y = asscaledarray(x)
    >>> x is y
    False
    >>> np.asarray(y).round(3)
    array([ 1.25 ,  8.125,  6.75 ,  4.625,  9.25 ])
    >>> z = asscaledarray(y)
    >>> y is z
    True
    >>> q = asscaledarray(y, forceoffset = 1.3)
    >>> z is q
    False
    >>> np.asarray(y).round(3)
    array([ 1.25 ,  8.125,  6.75 ,  4.625,  9.25 ])
    >>> z.offset
    5.25
    >>> q.offset
    1.3

    """
    if isinstance(x, scaledarray) and (
            forceoffset is None or forceoffset == x.offset):
        # nothing to do
        return x
    else:
        x = np.asarray(x)

        if len(x) == 0:
            if forceoffset is None:
                forceoffset = 0.
            return scaledarray(np.array([], dtype=np.int16), 1., forceoffset)
        elif len(x) == 1:
            if forceoffset is None:
                forceoffset = 0.
            return scaledarray(np.array([1], dtype=np.int16),
                    x[0] - forceoffset, forceoffset)
        else:
            if forceoffset is None:
                forceoffset = (x.max() + x.min())/2
            scale = np.max(np.abs(x - forceoffset)) / 32767

            return scaledarray(
                np.array((x - forceoffset) / scale).astype(np.int16),
                scale, forceoffset)



cdef prepare_columndata(ColumnData* columndata, colnum, fileformat, name, data):
    """Use the data in a python sequence to fill out a C ColumnData struct"""
    cdef np.ndarray[np.int16_t, ndim=1] shortdata
    cdef np.ndarray[np.int32_t, ndim=1] intdata
    cdef np.ndarray[np.float32_t, ndim=1] floatdata
    cdef np.ndarray[np.float64_t, ndim=1] doubledata

    memset(columndata, 0, sizeof(columndata))

    # fill in the column name
    columndata.titleLength = 2*len(name)
    columndata.title = <unsigned char*>malloc(columndata.titleLength+2)
    memcpy(columndata.title, <char*>name, len(name))
    columndata.title[len(name)] = 0

    # fill in the number of data points
    columndata.points = len(data)

    # down-convert the data to floats for the older axograph formats
    # (since they only support floats)
    if fileformat == old_digitized_format:
        if colnum == 0:
            data = aslinearsequence(data)
        else:
            data = asscaledarray(data, forceoffset = 0.)
    elif fileformat == old_graph_format:
        data = np.asarray(data, dtype=np.float32)

    # create a column of the appropriate type
    if isinstance(data, linearsequence):
        columndata.type = SeriesArrayType
        columndata.seriesArray.firstValue = data.start
        columndata.seriesArray.increment = data.step
    elif isinstance(data, scaledarray):
        columndata.type = ScaledShortArrayType
        columndata.scaledShortArray.scale = data.scale
        columndata.scaledShortArray.offset = data.offset
        shortdata = data.data
        columndata.scaledShortArray.shortArray = <short*>malloc(
                columndata.points * sizeof(short))
        memcpy(columndata.scaledShortArray.shortArray,
                <short*>shortdata.data,
                columndata.points * sizeof(short))
    else:
        # convert it to an array for further processing
        array = np.asarray(data)

        if array.dtype == np.int16:
            columndata.type = ShortArrayType
            shortdata = array
            columndata.shortArray = <short*>malloc(
                    columndata.points * sizeof(short))
            memcpy(columndata.shortArray, <short*>shortdata.data,
                    columndata.points * sizeof(short))
        elif array.dtype == np.int32:
            columndata.type = IntArrayType
            intdata = array
            columndata.intArray = <int32_t*>malloc(
                    columndata.points * sizeof(int32_t))
            memcpy(columndata.intArray, <int32_t*>intdata.data,
                    columndata.points * sizeof(int32_t))
        elif array.dtype == np.float32:
            columndata.type = FloatArrayType
            floatdata = array
            columndata.floatArray = <float*>malloc(
                    columndata.points * sizeof(float))
            memcpy(columndata.floatArray, <float*>floatdata.data,
                    columndata.points * sizeof(float))
        elif array.dtype == np.float64:
            columndata.type = DoubleArrayType
            doubledata = array
            columndata.doubleArray = <double*>malloc(
                    columndata.points * sizeof(double))
            memcpy(columndata.doubleArray, <double*>doubledata.data,
                    columndata.points * sizeof(double))
        else:
            raise TypeError("Unsupported column data type %s"
                    % repr(array.dtype))



cdef convert_columndata(ColumnData* columndata):
    """Convert the data in a C ColumnData struct to a python sequence"""
    cdef np.ndarray[np.int16_t, ndim=1] shortdata
    cdef np.ndarray[np.int32_t, ndim=1] intdata
    cdef np.ndarray[np.float32_t, ndim=1] floatdata
    cdef np.ndarray[np.float64_t, ndim=1] doubledata

    if columndata.type == ShortArrayType:
        shortdata = np.empty(columndata.points, dtype = np.int16)
        memcpy(shortdata.data, columndata.shortArray,
                columndata.points * sizeof(short))
        return shortdata
    elif columndata.type == IntArrayType:
        intdata = np.empty(columndata.points, dtype = np.int32)
        memcpy(intdata.data, columndata.intArray,
                columndata.points * sizeof(int32_t))
        return intdata
    elif columndata.type == FloatArrayType:
        floatdata = np.empty(columndata.points, dtype = np.float32)
        memcpy(floatdata.data, columndata.floatArray,
                columndata.points * sizeof(float))
        return floatdata
    elif columndata.type == DoubleArrayType:
        doubledata = np.empty(columndata.points, dtype = np.float64)
        memcpy(doubledata.data, columndata.doubleArray,
                columndata.points * sizeof(double))
        return doubledata
    elif columndata.type == SeriesArrayType:
        return linearsequence(columndata.points,
                columndata.seriesArray.firstValue,
                columndata.seriesArray.increment)
    elif columndata.type == ScaledShortArrayType:
        shortdata = np.empty(columndata.points, dtype = np.int16)
        memcpy(shortdata.data, columndata.scaledShortArray.shortArray,
                columndata.points * sizeof(short))
        return scaledarray(shortdata, columndata.scaledShortArray.scale,
            columndata.scaledShortArray.offset)
    else:
        raise IOError('Unsupported column type %d' % columndata.type)



cdef free_columndata(ColumnData* columndata):
    """Free the memory used by a columndata structure"""

    free(<char*>columndata.title)
    columndata.title = NULL
    columndata.titleLength = 0

    if columndata.type == ShortArrayType:
        free(columndata.shortArray)
        columndata.shortArray = NULL
    elif columndata.type == IntArrayType:
        free(columndata.intArray)
        columndata.intArray = NULL
    elif columndata.type == FloatArrayType:
        free(columndata.floatArray)
        columndata.floatArray = NULL
    elif columndata.type == DoubleArrayType:
        free(columndata.doubleArray)
        columndata.doubleArray = NULL
    elif columndata.type == SeriesArrayType:
        pass
    elif columndata.type == ScaledShortArrayType:
        free(columndata.scaledShortArray.shortArray)
        columndata.scaledShortArray.shortArray = NULL
    else:
        raise IOError('Unsupported column type %d' % columndata.type)



def read(char* filename):
    """Read an Axograph file

    Read an Axograph file from disk and return the contents as an
    axographio.file_contents object.

    """
    cdef int fileformat = 0
    cdef int result
    cdef int32_t numcolumns
    cdef ColumnData columndata
    cdef unsigned int i

    # open the file
    cdef AGDataRef file = OpenFile(filename)
    if file == NULL:
        raise IOError('file not found')

    try:
        # figure out the file format
        result = AG_GetFileFormat( file, &fileformat )
        if result != 0:
            if result == kAG_FormatErr:
                raise IOError('file is not in AxoGraph format')
            elif result == kAG_VersionErr:
                raise IOError('file is not in AxoGraph format')
            else:
                raise IOError((result,
                    'AG_GetFileFormat returned error %d' % result))

        # read in the number of columns
        result = AG_GetNumberOfColumns(file, fileformat, &numcolumns)
        if result != 0:
            raise IOError((result,
                'AG_GetNumberOfColumns returned error %d' % result))
        elif numcolumns < 0:
            raise IOError('number of columns was negative')

        # read in each column of data
        colnames = []
        coldata = []
        for colnum in range(numcolumns):
            result = AG_ReadColumn(file, fileformat, colnum, &columndata)
            if result != 0:
                raise IOError((result,
                    'AG_ReadColumn returned error %d' % result))

            if <char*>columndata.title is None:
                colname = '' #'Column %d' % colnum
            else:
                colname = <char*>(columndata.title)
            colnames += [colname]
            coldata += [convert_columndata(&columndata)]
            free_columndata(&columndata)

    finally:
        CloseFile(file)

    return file_contents(colnames, coldata, fileformat)



"""A Python package for reading and writing AxoGraph data files

Use `f = axographio.read(filename)` to load an AxoGraph file. Data and column
names are accessed as lists with `f.data` and `f.names`. Data objects (with
types `axographio.extension.linearsequence` and
`axographio.extension.scaledarray`) can be converted to NumPy or SciPy arrays,
e.g., `numpy.asarray(f.data[0])`.

Tests can be run using `axographio.tests.run()`.

To get started, try executing this code, which uses an example AxoGraph data
file included with the axographio package:

>>> import axographio
>>> import matplotlib.pyplot as plt
>>>
>>> f = axographio.read(axographio.example_files['axograph_x_format'])
>>>
>>> plt.plot(f.data[0], f.data[1])
>>> plt.xlabel(f.names[0])
>>> plt.ylabel(f.names[1])
>>> plt.show()  # may be optional depending on your OS
"""

from .version import version as __version__
from .version import git_revision as __git_revision__
from .extension import *
from . import tests
from .tests.test_axographio import example_files

# explicit listing needed to ensure pydoc/help() finds everything
__all__ = [
    'file_contents',
    'linearsequence',
    'scaledarray',
    'aslinearsequence',
    'asscaledarray',
    'read',
    'axograph_x_format',
    'newest_format',
    'old_digitized_format',
    'old_graph_format',
    'supported_formats',
    'example_files',
    ]



#!/usr/bin/env python

import os
import sys
import subprocess
from setuptools import setup, Extension
import numpy


# Change version number here, not in axographio/version.py, which is generated
# by this script. Try to follow recommended versioning guidelines at semver.org.
MAJOR       = 0     # increment for backwards-incompatible changes
MINOR       = 3     # increment for backwards-compatible feature additions
MICRO       = 2     # increment for backwards-compatible bug fixes
IS_RELEASED = True  # determines whether version will be marked as development
VERSION     = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


# This function may be used below to query git for a revision number.
def _minimal_ext_cmd(cmd):
    """Run an external command and return the result."""
    # construct minimal environment
    env = {}
    for k in ['SYSTEMROOT', 'PATH', 'HOME']:
        v = os.environ.get(k)
        if v is not None:
            env[k] = v
    # LANGUAGE is used on win32
    env['LANGUAGE'] = 'C'
    env['LANG'] = 'C'
    env['LC_ALL'] = 'C'
    out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
    return out


# Try to fetch the git revision number from the .git directory if it exists.
if os.path.exists(".git"):
    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "unknown"
# If the .git directory is absent (perhaps because this is a source distro),
# try to fetch the rev number from axographio/version.py where it may have been
# stored during packaging.
elif os.path.exists("axographio/version.py"):
    try:
        v = {}
        with open("axographio/version.py", "r") as f:
            exec(f.read(), v)
        GIT_REVISION = v['git_revision']
    except ImportError:
        raise ImportError("Unable to import git_revision. Try removing " \
                          "axographio/version.py and the build directory " \
                          "before building.")
else:
    GIT_REVISION = "unknown"


# If this is not a release version, mark it as a development build/distro and
# tag it with the git revision number.
if not IS_RELEASED:
    VERSION += '.dev0+' + GIT_REVISION[:7]


# Write the version string to a file that will be included with the
# build/distro. This makes the string accessible to the package via
# axographio.__version__. The git revision is also written in case a source
# distro is being built, so that it can be fetched later during installation.
with open("axographio/version.py", "w") as f:
    try:
        f.write("\"\"\"THIS FILE WAS GENERATED BY SETUP.PY DURING BUILDING/PACKAGING\"\"\"\n")
        f.write("version = '%s'\n" % VERSION)
        f.write("git_revision = '%s'\n" % GIT_REVISION)
    finally:
        f.close()


# Read in the README to serve as the long_description, which will be presented
# on pypi.org as the project description.
with open("README.rst", "r") as f:
    README = f.read()


setup(
    name = "axographio",
    version = VERSION,
    setup_requires   = ['numpy', 'cython>=0.19'],   # needed to build axographio
    install_requires = ['numpy'],                   # needed to run axographio
    packages = ['axographio', 'axographio.tests'],
    ext_modules = [
        Extension('axographio.extension', [
            'axographio/extension.pyx',
            'axographio/include/axograph_readwrite/fileUtils.cpp',
            'axographio/include/axograph_readwrite/byteswap.cpp',
            'axographio/include/axograph_readwrite/stringUtils.cpp',
            'axographio/include/axograph_readwrite/AxoGraph_ReadWrite.cpp'],
            language='c++', include_dirs=[numpy.get_include()],
            define_macros=[('NO_CARBON',1)]
            )
        ],
    test_suite = 'axographio.tests.test_axographio.test_suite',
    package_data = {'axographio.tests': [
        '../include/axograph_readwrite/AxoGraph Digitized File',
        '../include/axograph_readwrite/AxoGraph Graph File',
        '../include/axograph_readwrite/AxoGraph X File.axgx'
        ]},
    # metatdata
    author = "Kendrick Shaw, Jeffrey Gill",
    author_email = "kms15@case.edu, jeffrey.p.gill@gmail.com",
    maintainer = "Jeffrey Gill",
    maintainer_email = "jeffrey.p.gill@gmail.com",
    license = "BSD License",
    keywords = ["physiology","electrophysiology","axograph"],
    url = "https://github.com/CWRUChielLab/axographio",
    description = "A Python package for reading and writing AxoGraph data files",
    long_description = README,
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
        ]
    )
